package SEU.centralized;

import adf.agent.action.Action;
import adf.agent.action.common.ActionMove;
import adf.agent.action.common.ActionRest;
import adf.agent.communication.MessageManager;
import adf.agent.communication.standard.bundle.centralized.CommandAmbulance;
import adf.agent.communication.standard.bundle.centralized.CommandPolice;
import adf.agent.communication.standard.bundle.centralized.MessageReport;
import adf.agent.communication.standard.bundle.information.MessagePoliceForce;
import adf.agent.communication.standard.bundle.information.MessageRoad;
import adf.agent.develop.DevelopData;
import adf.agent.info.AgentInfo;
import adf.agent.info.ScenarioInfo;
import adf.agent.info.WorldInfo;
import adf.agent.module.ModuleManager;
import adf.agent.precompute.PrecomputeData;
import adf.component.centralized.CommandExecutor;
import adf.component.communication.CommunicationMessage;
import adf.component.extaction.ExtAction;
import adf.component.module.algorithm.PathPlanning;
import rescuecore2.misc.geometry.GeometryTools2D;
import rescuecore2.misc.geometry.Line2D;
import rescuecore2.misc.geometry.Point2D;
import rescuecore2.misc.geometry.Vector2D;
import rescuecore2.standard.entities.*;
import rescuecore2.worldmodel.EntityID;

import java.util.*;

import static rescuecore2.standard.entities.StandardEntityURN.BLOCKADE;
import static rescuecore2.standard.entities.StandardEntityURN.REFUGE;

public class CommandExecutorPolice extends CommandExecutor<CommandPolice> {
    private static final int ACTION_UNKNOWN = -1;
    private static final int ACTION_REST = CommandPolice.ACTION_REST;
    private static final int ACTION_MOVE = CommandPolice.ACTION_MOVE;
    private static final int ACTION_CLEAR = CommandPolice.ACTION_CLEAR;
    private static final int ACTION_AUTONOMY = CommandPolice.ACTION_AUTONOMY;

    //command
    private int commandType;
    private EntityID target;
    private EntityID commanderID;

    //tool
    private PathPlanning pathPlanning;
    private ExtAction actionExtClear;
    private ExtAction actionExtMove;

    //history
    private Set<EntityID> clearedAreas;
    private int lastRequestTime = -5;

    //observable entities
    private Set<EntityID> stuckAgentRoads;

    //constant
    private final int stuckRange = 600;
    private final int sendingAvoidTimeRequest = 5;

    //precompute key
    private static final String KEY_CLEARED_AREAS = "sample.commandExecutorPolice.ClearedAreas";

    public CommandExecutorPolice(AgentInfo ai, WorldInfo wi, ScenarioInfo si, ModuleManager moduleManager, DevelopData developData) {
        super(ai, wi, si, moduleManager, developData);
        this.commandType = ACTION_UNKNOWN;

        switch  (scenarioInfo.getMode()) {
            case PRECOMPUTATION_PHASE:
                this.pathPlanning = moduleManager.getModule("CommandExecutorPolice.PathPlanning", "adf.sample.module.algorithm.SamplePathPlanning");
                this.actionExtClear = moduleManager.getExtAction("CommandExecutorPolice.ActionExtClear", "adf.sample.extaction.ActionExtClear");
                this.actionExtMove = moduleManager.getExtAction("CommandExecutorPolice.ActionExtMove", "adf.sample.extaction.ActionExtMove");
                break;
            case PRECOMPUTED:
                this.pathPlanning = moduleManager.getModule("CommandExecutorPolice.PathPlanning", "adf.sample.module.algorithm.SamplePathPlanning");
                this.actionExtClear = moduleManager.getExtAction("CommandExecutorPolice.ActionExtClear", "adf.sample.extaction.ActionExtClear");
                this.actionExtMove = moduleManager.getExtAction("CommandExecutorPolice.ActionExtMove", "adf.sample.extaction.ActionExtMove");
                break;
            case NON_PRECOMPUTE:
                this.pathPlanning = moduleManager.getModule("CommandExecutorPolice.PathPlanning", "adf.sample.module.algorithm.SamplePathPlanning");
                this.actionExtClear = moduleManager.getExtAction("CommandExecutorPolice.ActionExtClear", "adf.sample.extaction.ActionExtClear");
                this.actionExtMove = moduleManager.getExtAction("CommandExecutorPolice.ActionExtMove", "adf.sample.extaction.ActionExtMove");
                break;
        }

        this.clearedAreas = new HashSet<>();
        this.stuckAgentRoads = new HashSet<>();
    }

    @Override
    public CommandExecutor setCommand(CommandPolice command) {
        EntityID agentID = this.agentInfo.getID();
        if(command.isToIDDefined() && Objects.requireNonNull(command.getToID()).getValue() == agentID.getValue()) {
            this.commandType = command.getAction();
            this.target = command.getTargetID();
            this.commanderID = command.getSenderID();
        }
        return this;
    }

    public CommandExecutor precompute(PrecomputeData precomputeData) {
        super.precompute(precomputeData);
        if(this.getCountPrecompute() >= 2) {
            return this;
        }
        this.pathPlanning.precompute(precomputeData);
        this.actionExtClear.precompute(precomputeData);
        this.actionExtMove.precompute(precomputeData);

        Queue<EntityID> clearedQueue = new ArrayDeque<>();
        for (StandardEntity entity : this.worldInfo.getEntitiesOfType(StandardEntityURN.ROAD)) {
            if (((Area)entity).getNeighbours().size() <= 1) {
                clearedQueue.offer(entity.getID());
            }
        }
        while (!clearedQueue.isEmpty()) {
            EntityID clearedID = clearedQueue.poll();
            this.clearedAreas.add(clearedID);
            for (EntityID id : ((Area)this.worldInfo.getEntity(clearedID)).getNeighbours()) {
                if (clearedQueue.contains(id) || this.clearedAreas.contains(id)) {
                    continue;
                }
                Area area = (Area)this.worldInfo.getEntity(id);
                int count = 0;
                for (EntityID neighbor : area.getNeighbours()) {
                    if (this.clearedAreas.contains(neighbor) || clearedQueue.contains(neighbor)) {
                        ++count;
                    }
                }
                if (area.getNeighbours().size() - count <= 1) {
                    clearedQueue.offer(id);
                }
            }
        }

        precomputeData.setEntityIDList(KEY_CLEARED_AREAS, new ArrayList<>(this.clearedAreas));

        return this;
    }

    public CommandExecutor resume(PrecomputeData precomputeData) {
        super.resume(precomputeData);
        if(this.getCountResume() >= 2) {
            return this;
        }
        this.pathPlanning.resume(precomputeData);
        this.actionExtClear.resume(precomputeData);
        this.actionExtMove.resume(precomputeData);

        this.clearedAreas = new HashSet<>(precomputeData.getEntityIDList(KEY_CLEARED_AREAS));

        return this;
    }

    public CommandExecutor preparate() {
        super.preparate();
        if(this.getCountPreparate() >= 2) {
            return this;
        }
        this.pathPlanning.preparate();
        this.actionExtClear.preparate();
        this.actionExtMove.preparate();

        Queue<EntityID> clearedQueue = new ArrayDeque<>();
        for (StandardEntity entity : this.worldInfo.getEntitiesOfType(StandardEntityURN.ROAD)) {
            if (((Area)entity).getNeighbours().size() <= 1) {
                clearedQueue.offer(entity.getID());
            }
        }
        while (!clearedQueue.isEmpty()) {
            EntityID clearedID = clearedQueue.poll();
            this.clearedAreas.add(clearedID);
            for (EntityID id : ((Area)this.worldInfo.getEntity(clearedID)).getNeighbours()) {
                if (clearedQueue.contains(id) || this.clearedAreas.contains(id)) {
                    continue;
                }
                Area area = (Area)this.worldInfo.getEntity(id);
                int count = 0;
                for (EntityID neighbor : area.getNeighbours()) {
                    if (this.clearedAreas.contains(neighbor) || clearedQueue.contains(neighbor)) {
                        ++count;
                    }
                }
                if (area.getNeighbours().size() - count <= 1) {
                    clearedQueue.offer(id);
                }
            }
        }

        return this;
    }

    public CommandExecutor updateInfo(MessageManager messageManager){
        super.updateInfo(messageManager);
        if(this.getCountUpdateInfo() >= 2) {
            return this;
        }
        this.pathPlanning.updateInfo(messageManager);
        this.actionExtClear.updateInfo(messageManager);
        this.actionExtMove.updateInfo(messageManager);

        this.sendRequestCommand(messageManager);

        this.updateChanged();

        this.reflectMessage(messageManager);

        if(this.isCommandCompleted()) {
            if(this.commandType != ACTION_UNKNOWN) {
                messageManager.addMessage(new MessageReport(true, true, true, this.commanderID));
                this.commandType = ACTION_UNKNOWN;
                this.target = null;
                this.commanderID = null;
            }
        }

        return this;
    }

    @Override
    public CommandExecutor calc() {
        this.result = null;
        EntityID position = this.agentInfo.getPosition();
        switch (this.commandType) {
            case ACTION_REST:
                if(this.target == null) {
                    if(worldInfo.getEntity(position).getStandardURN() != REFUGE) {
                        this.pathPlanning.setFrom(position);
                        this.pathPlanning.setDestination(this.worldInfo.getEntityIDsOfType(REFUGE));
                        List<EntityID> path = this.pathPlanning.calc().getResult();
                        if(path != null && path.size() > 0) {
                            Action action = this.actionExtClear.setTarget(path.get(path.size() - 1)).calc().getAction();
                            if(action == null) {
                                action = new ActionMove(path);
                            }
                            this.result = action;
                            return this;
                        }
                    }
                } else if (position.getValue() != this.target.getValue()) {
                    List<EntityID> path = this.pathPlanning.getResult(position, this.target);
                    if(path != null && path.size() > 0) {
                        Action action = this.actionExtClear.setTarget(path.get(path.size() - 1)).calc().getAction();
                        if(action == null) {
                            action = new ActionMove(path);
                        }
                        this.result = action;
                        return this;
                    }
                }
                this.result = new ActionRest();
                return this;
            case ACTION_MOVE:
                if(this.target != null) {
                    this.result = this.actionExtClear.setTarget(this.target).calc().getAction();
                }
                return this;
            case ACTION_CLEAR:
                if(this.target != null) {
                    this.result = this.actionExtClear.setTarget(this.target).calc().getAction();
                }
                return this;
            case ACTION_AUTONOMY:
                if(this.target == null) {
                    return this;
                }
                StandardEntity targetEntity = this.worldInfo.getEntity(this.target);
                if(targetEntity.getStandardURN() == REFUGE) {
                    PoliceForce agent = (PoliceForce) this.agentInfo.me();
                    if(agent.getDamage() > 0) {
                        if (position.getValue() != this.target.getValue()) {
                            List<EntityID> path = this.pathPlanning.getResult(position, this.target);
                            if(path != null && path.size() > 0) {
                                Action action = this.actionExtClear.setTarget(path.get(path.size() - 1)).calc().getAction();
                                if (action == null) {
                                    action = new ActionMove(path);
                                }
                                this.result = action;
                                return this;
                            }
                        }
                        this.result = new ActionRest();
                    } else {
                        this.result = this.actionExtClear.setTarget(this.target).calc().getAction();
                    }
                } else if (targetEntity instanceof Area) {
                    this.result = this.actionExtClear.setTarget(this.target).calc().getAction();
                    return this;
                }else if (targetEntity instanceof Human) {
                    Human h = (Human) targetEntity;
                    if((h.isHPDefined() && h.getHP() == 0)) {
                        return this;
                    }
                    if(h.isPositionDefined() && this.worldInfo.getPosition(h) instanceof Area) {
                        this.target = h.getPosition();
                        this.result = this.actionExtClear.setTarget(this.target).calc().getAction();
                    }
                } else if(targetEntity.getStandardURN() == BLOCKADE) {
                    Blockade blockade = (Blockade)targetEntity;
                    if(blockade.isPositionDefined()) {
                        this.target = blockade.getPosition();
                        this.result = this.actionExtClear.setTarget(this.target).calc().getAction();
                    }
                }
        }
        return this;
    }

    //UpdateInfo
    private void sendRequestCommand(MessageManager messageManager) {
        if (this.agentInfo.getTime() < this.lastRequestTime + this.sendingAvoidTimeRequest) {
            return;
        }
        Human me = (Human)this.agentInfo.me();
        if (me.isBuriednessDefined() && me.getBuriedness() > 0) {
            messageManager.addMessage(new CommandAmbulance(true, null, this.agentInfo.getPosition(), CommandAmbulance.ACTION_MOVE));
            Area position = (Area)this.worldInfo.getPosition(this.agentInfo.getID());
            if (position instanceof Building) {
                Building building = (Building)position;
                EntityID entrance = this.getClosestEntityID(this.getAllEntrancesOfBuilding(building), this.agentInfo.getID());
                if (!this.isPassable((Area)this.worldInfo.getEntity(entrance))) {
                    messageManager.addMessage(new CommandPolice(true, null, entrance, CommandPolice.ACTION_CLEAR));
                }
            }
        }
    }

    private void updateChanged() {
        this.stuckAgentRoads.clear();
        for (EntityID id : this.worldInfo.getChanged().getChangedEntities()) {
            StandardEntity entity = this.worldInfo.getEntity(id);
            if ((entity instanceof FireBrigade) || (entity instanceof AmbulanceTeam) || (entity instanceof Civilian)) {
                Human human = (Human)entity;
                EntityID position = human.getPosition();
                StandardEntity positionEntity = this.worldInfo.getEntity(position);
                if (positionEntity instanceof Road) {
                    if (this.isHumanStucked(human, (Road)positionEntity)) {
                        this.stuckAgentRoads.add(position);
                    }
                } else if (positionEntity instanceof Building) {
                    boolean flag = false;
                    for (EntityID entrance : this.getAllEntrancesOfBuilding((Building)positionEntity)) {
                        if (this.isPassable((Area)this.worldInfo.getEntity(entrance))) {
                            flag = true;
                            break;
                        }
                    }
                    if (!flag) {
                        this.stuckAgentRoads.addAll(this.getAllEntrancesOfBuilding((Building)positionEntity));
                    }
                }
            } else if (entity instanceof Road) {
                if (this.isPassable((Road)entity)) {
                    this.clearedAreas.add(id);
                }
            }
        }
    }

    private void reflectMessage(MessageManager messageManager) {
        Collection<EntityID> change = this.worldInfo.getChanged().getChangedEntities();
        //reflect MessageRoad
        for (CommunicationMessage message : messageManager.getReceivedMessageList(MessageRoad.class)) {
            MessageRoad messageRoad = (MessageRoad)message;
            EntityID roadID = messageRoad.getRoadID();
            if (!change.contains(roadID) && messageRoad.isPassable()) {
                this.clearedAreas.add(roadID);
            }
        }
        //reflect MessagePoliceForce
        for (CommunicationMessage message : messageManager.getReceivedMessageList(MessagePoliceForce.class)) {
            MessagePoliceForce messagePoliceForce = (MessagePoliceForce)message;
            if (messagePoliceForce.isPositionDefined() && !change.contains(messagePoliceForce.getPosition())) {
                this.clearedAreas.add(messagePoliceForce.getPosition());
            }
            if (messagePoliceForce.isTargetDefined() && !change.contains(messagePoliceForce.getTargetID())
                    && this.target != null && !this.target.equals(messagePoliceForce.getTargetID())) {
                this.clearedAreas.add(messagePoliceForce.getTargetID());
            }
        }
    }

    private boolean isCommandCompleted() {
        PoliceForce agent = (PoliceForce) this.agentInfo.me();
        switch (this.commandType) {
            case ACTION_REST:
                if(this.target == null) {
                    return (agent.getDamage() == 0);
                }
                if(this.worldInfo.getEntity(this.target).getStandardURN() == REFUGE) {
                    if (agent.getPosition().getValue() == this.target.getValue()) {
                        return (agent.getDamage() == 0);
                    }
                }
                return false;
            case ACTION_MOVE:
                return this.target == null || (this.agentInfo.getPosition().getValue() == this.target.getValue());
            case ACTION_CLEAR:
                if(this.target == null) {
                    return true;
                }
                StandardEntity entity = this.worldInfo.getEntity(this.target);
                if(entity instanceof Road) {
                    if (this.clearedAreas.contains(this.target) && !this.stuckAgentRoads.contains(this.target)) {
                        return true;
                    }
                    return false;
                }
                return true;
            case ACTION_AUTONOMY:
                if(this.target != null) {
                    StandardEntity targetEntity = this.worldInfo.getEntity(this.target);
                    if(targetEntity.getStandardURN() == REFUGE) {
                        this.commandType = agent.getDamage() > 0 ? ACTION_REST : ACTION_CLEAR;
                        return this.isCommandCompleted();
                    } else if (targetEntity instanceof Area) {
                        this.commandType = ACTION_CLEAR;
                        return this.isCommandCompleted();
                    }else if (targetEntity instanceof Human) {
                        Human h = (Human) targetEntity;
                        if((h.isHPDefined() && h.getHP() == 0)) {
                            return true;
                        }
                        if(h.isPositionDefined() && this.worldInfo.getPosition(h) instanceof Area) {
                            this.target = h.getPosition();
                            this.commandType = ACTION_CLEAR;
                            return this.isCommandCompleted();
                        }
                    } else if(targetEntity.getStandardURN() == BLOCKADE) {
                        Blockade blockade = (Blockade)targetEntity;
                        if(blockade.isPositionDefined()) {
                            this.target = blockade.getPosition();
                            this.commandType = ACTION_CLEAR;
                            return this.isCommandCompleted();
                        }
                    }
                }
                return true;
        }
        return true;
    }

    //Entity Utils
    private boolean isHumanStucked(Human human, Road road) {
        if (!road.isBlockadesDefined() || road.getBlockades().isEmpty()) {
            return false;
        }
        for (Blockade blockade : this.worldInfo.getBlockades(road)) {
            int hX = human.getX();
            int hY = human.getY();
            if (this.isInside(hX, hY, blockade.getApexes()) ||
                    this.isNearBlockade(hX, hY, blockade, this.stuckRange)) {
                return true;
            }
        }
        return false;
    }

    private boolean isPassable(Area area) {
        if (!(area instanceof Road)) {
            return true;
        }
        if (!area.isBlockadesDefined() || area.getBlockades().isEmpty()) {
            return true;
        }
        Collection<Blockade> blockades = this.worldInfo.getBlockades(area);
        int X = area.getX();
        int Y = area.getY();
        for (Edge edge : this.getPassableEdge(area)) {
            Point2D mid = this.getMidPoint(edge.getLine());
            double midX = mid.getX();
            double midY = mid.getY();
            Vector2D vector = new Vector2D(X - mid.getX(), Y - mid.getY());
            vector = vector.normalised().scale(250).getNormal();
            for (Blockade blockade : blockades) {
                if (blockade.isApexesDefined()) {
                    if (this.intersect(X, Y, midX, midY, blockade) ||
                            this.intersect(X + vector.getX(), Y + vector.getY(),
                                    midX + vector.getX(), midY + vector.getY(), blockade) ||
                            this.intersect(X - vector.getX(), Y - vector.getY(),
                                    midX - vector.getX(), midY - vector.getY(), blockade)) {
                        return false;
                    }
                }
            }
        }
        return true;
    }

    private Set<Edge> getPassableEdge(Area area) {
        Set<Edge> passableEdges = new HashSet<>();
        for (Edge edge : area.getEdges()) {
            if (edge.isPassable()) {
                passableEdges.add(edge);
            }
        }
        return passableEdges;
    }

    private Set<EntityID> getAllEntrancesOfBuilding(Building building) {
        EntityID buildingID = building.getID();
        Set<EntityID> entrances = new HashSet<>();
        Set<EntityID> visited = new HashSet<>();
        Stack<EntityID> stack = new Stack<>();
        stack.push(buildingID);
        visited.add(buildingID);
        while (!stack.isEmpty()) {
            EntityID id = stack.pop();
            visited.add(id);
            Area area = (Area)this.worldInfo.getEntity(id);
            if (area instanceof Road) {
                entrances.add(id);
            } else {
                for (EntityID neighbor : area.getNeighbours()) {
                    if (visited.contains(neighbor)) {
                        continue;
                    }
                    stack.push(neighbor);
                }
            }
        }
        return entrances;
    }

    //Geom Utils
    private Point2D getMidPoint(Line2D line) {
        return new Point2D((line.getOrigin().getX() + line.getEndPoint().getX()) / 2,
                (line.getOrigin().getY() + line.getEndPoint().getY()) / 2);
    }

    private boolean intersect(double agentX, double agentY, double pointX, double pointY, Blockade blockade) {
        List<Line2D> lines = GeometryTools2D.pointsToLines(GeometryTools2D.vertexArrayToPoints(blockade.getApexes()), true);
        for(Line2D line : lines) {
            Point2D start = line.getOrigin();
            Point2D end = line.getEndPoint();
            double startX = start.getX();
            double startY = start.getY();
            double endX = end.getX();
            double endY = end.getY();
            if(java.awt.geom.Line2D.linesIntersect(
                    agentX, agentY, pointX, pointY,
                    startX, startY, endX, endY
            )) {
                return true;
            }
        }
        return false;
    }

    private boolean isInside(double pX, double pY, int[] apex) {
        Point2D p = new Point2D(pX, pY);
        Vector2D v1 = (new Point2D(apex[apex.length - 2], apex[apex.length - 1])).minus(p);
        Vector2D v2 = (new Point2D(apex[0], apex[1])).minus(p);
        double theta = this.getAngle(v1, v2);

        for(int i = 0; i < apex.length - 2; i += 2) {
            v1 = (new Point2D(apex[i], apex[i + 1])).minus(p);
            v2 = (new Point2D(apex[i + 2], apex[i + 3])).minus(p);
            theta += this.getAngle(v1, v2);
        }
        return Math.round(Math.abs((theta / 2) / Math.PI)) >= 1;
    }

    private boolean isNearBlockade(double pX, double pY, Blockade blockade, double range) {
        int[] apex = blockade.getApexes();
        for (int i = 0; i < apex.length - 4; i += 2) {
            if(java.awt.geom.Line2D.ptLineDist(apex[i], apex[i + 1], apex[i + 2], apex[i + 3], pX, pY) < range) {
                return true;
            }
        }
        if (java.awt.geom.Line2D.ptLineDist(apex[0], apex[1], apex[apex.length - 2], apex[apex.length - 1], pX, pY) < range) {
            return true;
        }
        return false;
    }

    private double getAngle(Vector2D v1, Vector2D v2) {
        double flag = (v1.getX() * v2.getY()) - (v1.getY() * v2.getX());
        double angle = Math.acos(((v1.getX() * v2.getX()) + (v1.getY() * v2.getY())) / (v1.getLength() * v2.getLength()));
        if(flag > 0) {
            return angle;
        }
        if(flag < 0) {
            return -1 * angle;
        }
        return 0.0D;
    }

    //Selector
    private EntityID getClosestEntityID(Collection<EntityID> IDs, EntityID reference) {
        if (IDs.isEmpty()) {
            return null;
        }
        double minDistance = Double.MAX_VALUE;
        EntityID closestID = null;
        for (EntityID id : IDs) {
            double distance = this.worldInfo.getDistance(id, reference);
            if (distance < minDistance) {
                minDistance = distance;
                closestID = id;
            }
        }
        return closestID;
    }
}